@@ -30,7 +30,7 @@ static void
 set_string(char **strp, const char *value)
 {
 	if (*strp)
-		free(strp);
+		free(*strp);
 	*strp = value? strdup(value) : NULL;
 }
 
